USE AdventureWorks2017;
GO
/*	This script contains:
		1. A table to store performance counter data for long-term analytics.
		2. A table to store temporary data for cumulative counters.
		3. A stored procedure that takes a snapshot of current performance counter data.

	A job/task may be created that regularly calls the stored procedure, providing an automated collection
	process for populating this data.
*/

IF NOT EXISTS (SELECT * FROM sys.tables WHERE tables.name = 'performance_counters')
BEGIN
	CREATE TABLE dbo.performance_counters
	(	sample_time_utc DATETIME2(3) NOT NULL,
		page_life_expectancy_seconds BIGINT NOT NULL,
		page_reads_per_second DECIMAL(18,4) NOT NULL,
		page_writes_per_second DECIMAL(18,4) NOT NULL,
		lazy_writes_per_second DECIMAL(18,4) NOT NULL,
		page_lookups_per_second DECIMAL(18,4) NOT NULL,
		target_pages BIGINT NOT NULL,
		logins_per_second DECIMAL(18,4) NOT NULL,
		logouts_per_second DECIMAL(18,4) NOT NULL,
		temp_tables_created_per_second DECIMAL(18,4) NOT NULL,
		batch_requests_per_second DECIMAL(18,4) NOT NULL,
		compilations_per_second DECIMAL(18,4) NOT NULL,
		recompilations_per_second DECIMAL(18,4) NOT NULL);

	CREATE CLUSTERED COLUMNSTORE INDEX CCI_performance_counters ON dbo.performance_counters;
END
GO

IF NOT EXISTS (SELECT * FROM sys.tables WHERE tables.name = 'performance_counters_staging')
BEGIN
	CREATE TABLE dbo.performance_counters_staging
	(	sample_time_utc DATETIME2(3) NOT NULL,
		page_reads_cumulative DECIMAL(18,4) NOT NULL,
		page_writes_cumulative DECIMAL(18,4) NOT NULL,
		lazy_writes_cumulative DECIMAL(18,4) NOT NULL,
		page_lookups_cumulative DECIMAL(18,4) NOT NULL,
		logins_cumulative DECIMAL(18,4) NOT NULL,
		logouts_cumulative DECIMAL(18,4) NOT NULL,
		temp_tables_created_cumulative DECIMAL(18,4) NOT NULL,
		batch_requests_cumulative DECIMAL(18,4) NOT NULL,
		compilations_cumulative DECIMAL(18,4) NOT NULL,
		recompilations_cumulative DECIMAL(18,4) NOT NULL,
		CONSTRAINT PK_performance_counters PRIMARY KEY CLUSTERED (sample_time_utc));
END
GO

IF EXISTS (SELECT * FROM sys.procedures WHERE procedures.name = 'populate_performance_counters')
BEGIN
	DROP PROCEDURE dbo.populate_performance_counters;
END
GO

CREATE PROCEDURE dbo.populate_performance_counters
AS
BEGIN
	SET NOCOUNT ON;
	DECLARE @sample_time_utc DATETIME2(3) = GETUTCDATE();
	-- Put everything into a temp table to avoid having to hit dm_os_performance_counters multiple times.
	CREATE TABLE #performance_counters
	(	sample_time_utc DATETIME2(3) NOT NULL,
		page_life_expectancy_seconds BIGINT NOT NULL,
		page_reads_cumulative DECIMAL(18,4) NOT NULL,
		page_writes_cumulative DECIMAL(18,4) NOT NULL,
		lazy_writes_cumulative DECIMAL(18,4) NOT NULL,
		page_lookups_cumulative DECIMAL(18,4) NOT NULL,
		target_pages BIGINT NOT NULL,
		logins_cumulative DECIMAL(18,4) NOT NULL,
		logouts_cumulative DECIMAL(18,4) NOT NULL,
		temp_tables_created_cumulative DECIMAL(18,4) NOT NULL,
		batch_requests_cumulative DECIMAL(18,4) NOT NULL,
		compilations_cumulative DECIMAL(18,4) NOT NULL,
		recompilations_cumulative DECIMAL(18,4) NOT NULL);

	-- Collect all performance counters from dm_os_performance_counters
	INSERT INTO #performance_counters
		(sample_time_utc, page_life_expectancy_seconds, page_reads_cumulative, page_writes_cumulative,
	     lazy_writes_cumulative, page_lookups_cumulative, target_pages, logins_cumulative,
		 logouts_cumulative, temp_tables_created_cumulative, batch_requests_cumulative, compilations_cumulative,
		 recompilations_cumulative)
	SELECT
		@sample_time_utc,
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Page life expectancy' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS page_life_expectancy_seconds,	
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Page reads/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS page_reads_cumulative,	
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Page writes/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS page_writes_cumulative,	
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Lazy writes/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS lazy_writes_cumulative,	
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Page lookups/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS page_lookups_cumulative,	
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Target pages' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS target_pages,	
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Logins/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS logins_cumulative,
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Logouts/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS logouts_cumulative,
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Temp Tables Creation Rate' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS temp_tables_created_cumulative,
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'Batch Requests/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS batch_requests_cumulative,
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'SQL Compilations/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS compilations_cumulative,
		SUM(CASE WHEN dm_os_performance_counters.counter_name = 'SQL Re-Compilations/sec' THEN dm_os_performance_counters.cntr_value ELSE 0 END) AS recompilations_cumulative
	FROM sys.dm_os_performance_counters
	WHERE (dm_os_performance_counters.object_name LIKE '%Buffer Manager%' AND dm_os_performance_counters.counter_name IN ('Page life expectancy', 'Page reads/sec', 'Page writes/sec', 'Lazy writes/sec', 'Page lookups/sec', 'Target pages'))
	OR (dm_os_performance_counters.object_name LIKE '%General Statistics%' AND dm_os_performance_counters.counter_name IN ('Logins/sec', 'Logouts/sec', 'Temp Tables Creation Rate'))
	OR (dm_os_performance_counters.object_name LIKE '%SQL Statistics%' AND dm_os_performance_counters.counter_name IN ('Batch Requests/sec', 'SQL Compilations/sec', 'SQL Re-Compilations/sec'))

	-- Determine the last known entry of cumulative counters in the log and collect the sample time.
	DECLARE @last_sample_time datetime2(3);
	SELECT
		@last_sample_time = MAX(performance_counters_staging.sample_time_utc)
	FROM dbo.performance_counters_staging;

	-- Do not create performance data for the first run.  Instead, only seed the cumulative counters for the next run.
	IF @last_sample_time IS NOT NULL
	BEGIN
		DELETE performance_counters_staging
		FROM dbo.performance_counters_staging
		WHERE performance_counters_staging.sample_time_utc < DATEADD(DAY, -1, @last_sample_time);

		DECLARE @seconds_since_last_execution INT;
		SELECT @seconds_since_last_execution = DATEDIFF(SECOND, @last_sample_time, @sample_time_utc);

		-- Create perfmon data for the new point in time.
		INSERT INTO dbo.performance_counters
			(sample_time_utc, page_life_expectancy_seconds, page_reads_per_second, page_writes_per_second, lazy_writes_per_second,
			 page_lookups_per_second, target_pages, logins_per_second, logouts_per_second, temp_tables_created_per_second,
			 batch_requests_per_second, compilations_per_second, recompilations_per_second)
		SELECT
			performance_counters.sample_time_utc,
			performance_counters.page_life_expectancy_seconds,
			CASE WHEN performance_counters.page_reads_cumulative >= performance_counters_staging.page_reads_cumulative
				THEN performance_counters.page_reads_cumulative - performance_counters_staging.page_reads_cumulative
				ELSE performance_counters.page_reads_cumulative
			END / @seconds_since_last_execution AS page_reads_per_second,
			CASE WHEN performance_counters.page_writes_cumulative >= performance_counters_staging.page_writes_cumulative
				THEN performance_counters.page_writes_cumulative - performance_counters_staging.page_writes_cumulative
				ELSE performance_counters.page_writes_cumulative
			END / @seconds_since_last_execution AS page_writes_per_second,
			CASE WHEN performance_counters.lazy_writes_cumulative >= performance_counters_staging.lazy_writes_cumulative
				THEN performance_counters.lazy_writes_cumulative - performance_counters_staging.lazy_writes_cumulative
				ELSE performance_counters.lazy_writes_cumulative
			END / @seconds_since_last_execution AS lazy_writes_per_second,
			CASE WHEN performance_counters.page_lookups_cumulative >= performance_counters_staging.page_lookups_cumulative
				THEN performance_counters.page_lookups_cumulative - performance_counters_staging.page_lookups_cumulative
				ELSE performance_counters.page_lookups_cumulative
			END / @seconds_since_last_execution AS page_lookups_per_second,
			performance_counters.target_pages,
			CASE WHEN performance_counters.logins_cumulative >= performance_counters_staging.logins_cumulative
				THEN performance_counters.logins_cumulative - performance_counters_staging.logins_cumulative
				ELSE performance_counters.logins_cumulative
			END / @seconds_since_last_execution AS logins_per_second,
			CASE WHEN performance_counters.logouts_cumulative >= performance_counters_staging.logouts_cumulative
				THEN performance_counters.logouts_cumulative - performance_counters_staging.logouts_cumulative
				ELSE performance_counters.logouts_cumulative
			END / @seconds_since_last_execution AS logouts_per_second,
			CASE WHEN performance_counters.temp_tables_created_cumulative >= performance_counters_staging.temp_tables_created_cumulative
				THEN performance_counters.temp_tables_created_cumulative - performance_counters_staging.temp_tables_created_cumulative
				ELSE performance_counters.temp_tables_created_cumulative
			END / @seconds_since_last_execution AS temp_tables_created_per_second,
			CASE WHEN performance_counters.batch_requests_cumulative >= performance_counters_staging.batch_requests_cumulative
				THEN performance_counters.batch_requests_cumulative - performance_counters_staging.batch_requests_cumulative
				ELSE performance_counters.batch_requests_cumulative
			END / @seconds_since_last_execution AS batch_requests_per_second,
			CASE WHEN performance_counters.compilations_cumulative >= performance_counters_staging.compilations_cumulative
				THEN performance_counters.compilations_cumulative - performance_counters_staging.compilations_cumulative
				ELSE performance_counters.compilations_cumulative
			END / @seconds_since_last_execution AS compilations_per_second,
			CASE WHEN performance_counters.recompilations_cumulative >= performance_counters_staging.recompilations_cumulative
				THEN performance_counters.recompilations_cumulative - performance_counters_staging.recompilations_cumulative
				ELSE performance_counters.recompilations_cumulative
			END / @seconds_since_last_execution AS recompilations_per_second
		FROM #performance_counters performance_counters
		INNER JOIN dbo.performance_counters_staging -- This will skip inserting data for the first run as staging data did not exist yet
		ON performance_counters_staging.sample_time_utc = @last_sample_time;
	END

	-- Create perfmon data for the cumulative collectors that will be used for diffing minute-over-minute
	INSERT INTO dbo.performance_counters_staging
		(sample_time_utc, page_reads_cumulative, page_writes_cumulative, lazy_writes_cumulative, page_lookups_cumulative,
		 logins_cumulative, logouts_cumulative, temp_tables_created_cumulative, batch_requests_cumulative,
		 compilations_cumulative, recompilations_cumulative)
	SELECT
		sample_time_utc,
		page_reads_cumulative,
		page_writes_cumulative,
		lazy_writes_cumulative,
	    page_lookups_cumulative,
		logins_cumulative,
		logouts_cumulative,
		temp_tables_created_cumulative,
	    batch_requests_cumulative,
		compilations_cumulative,
		recompilations_cumulative
	FROM #performance_counters;

	DROP TABLE #performance_counters;
END
GO


EXEC dbo.populate_performance_counters;

SELECT * FROM dbo.performance_counters_staging;
SELECT * FROM dbo.performance_counters;
